@@ -1,25 +1,28 @@
-import asyncio
+import datetime
 import os
+import re
+import string
 import typing
-from cgi import parse_header
-from collections import namedtuple
 from tempfile import SpooledTemporaryFile
 from urllib.parse import SplitResult, parse_qsl, urlencode, urlsplit
 
-from .typing import Environ, Scope
+from .concurrency import run_in_threadpool
+from .typing import Environ, Final, Literal, Scope
+from .utils import parse_header
 
 __all__ = [
     "Address",
     "MediaType",
     "ContentType",
+    "Cookie",
     "URL",
     "MultiMapping",
     "MutableMultiMapping",
     "QueryParams",
-    "UploadFile",
-    "FormData",
     "Headers",
     "MutableHeaders",
+    "UploadFile",
+    "FormData",
 ]
 
 
@@ -28,7 +31,9 @@ KT = typing.TypeVar("KT")  # Key type.
 VT = typing.TypeVar("VT")  # Value type.
 
 
-Address = namedtuple("Address", ["host", "port"])
+class Address(typing.NamedTuple):
+    host: typing.Optional[str]
+    port: typing.Optional[int]
 
 
 class defaultdict(dict):
@@ -88,6 +93,100 @@ class ContentType:
         return self.type == other
 
 
+_cookie_legal_chars: Final[str] = (
+    string.ascii_letters + string.digits + "!#$%&'*+-.^_`|~:"
+)
+_cookie_is_legal_key = re.compile("[%s]+" % re.escape(_cookie_legal_chars)).fullmatch
+_cookie_translator: Final[typing.Dict[int, str]] = {
+    **{
+        n: "\\%03o" % n
+        for n in set(range(256)) - set(map(ord, _cookie_legal_chars + " ()/<=>?@[]{}"))
+    },
+    ord('"'): '\\"',
+    ord("\\"): "\\\\",
+}
+
+
+class Cookie:
+    def __init__(
+        self,
+        name: str,
+        value: str,
+        expires: typing.Optional[datetime.datetime] = None,
+        domain: typing.Optional[str] = None,
+        path: typing.Optional[str] = None,
+        httponly: bool = False,
+        secure: bool = False,
+        max_age: int = -1,
+        samesite: Literal["strict", "lax", "none"] = "lax",
+    ):
+        self.name = name
+        self.value = value
+        self.expires = expires
+        self.domain = domain
+        self.path = path
+        self.httponly = httponly
+        self.secure = secure
+        self.max_age = max_age
+        self.samesite = samesite
+
+    def _quote(self, value: str) -> str:
+        r"""
+        Quote a string for use in a cookie header.
+
+        If the string does not need to be double-quoted, then just return the
+        string.  Otherwise, surround the string in doublequotes and quote
+        (with a \) special characters.
+        """
+        if _cookie_is_legal_key(value):
+            return value
+        else:
+            return '"' + value.translate(_cookie_translator) + '"'
+
+    def __str__(self) -> str:
+        parts: typing.List[str] = []
+        parts.append(f"{self._quote(self.name)}={self._quote(self.value)}")
+
+        if self.expires:
+            parts.append(
+                "expires=" + self.expires.strftime("%a, %d %b %Y %H:%M:%S GMT")
+            )
+
+        if self.max_age > -1:
+            parts.append(f"max-age={self.max_age}")
+
+        if self.domain:
+            parts.append(f"domain={self.domain}")
+
+        if self.path:
+            parts.append(f"path={self.path}")
+
+        if self.httponly:
+            parts.append("httponly")
+
+        if self.secure or self.samesite in ("strict", "none"):
+            parts.append("secure")
+
+        parts.append(f"samesite={self.samesite}")
+
+        return "; ".join(parts)
+
+    def __bytes__(self) -> bytes:
+        return str(self).encode("ascii")
+
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, str):
+            return str(self) == other
+        if isinstance(other, bytes):
+            return bytes(self) == other
+        if isinstance(other, Cookie):
+            return str(self) == str(other)
+        return NotImplemented
+
+    def __repr__(self) -> str:
+        return f"<Cookie {self.name}: {self.value}>"
+
+
 class URL:
     __slots__ = ("_url", "_components")
 
@@ -95,55 +194,65 @@ class URL:
         self,
         url: str = "",
         *,
-        scope: Scope = None,
-        environ: Environ = None,
+        scope: typing.Optional[Scope] = None,
+        environ: typing.Optional[Environ] = None,
         **components: typing.Any,
     ) -> None:
-        if scope is not None or environ is not None:
-            assert not (url or components)
-            if scope is not None:
-                scheme = scope.get("scheme", "http")
-                server = scope.get("server", None)
-                path = scope.get("root_path", "") + scope["path"]
-                query_string = scope.get("query_string", b"")
-
-                host_header = None
-                for key, value in scope["headers"]:
-                    if key == b"host":
-                        host_header = value.decode("latin-1")
-                        break
-            elif environ is not None:
-                scheme = environ["wsgi.url_scheme"]
-                server = (environ["SERVER_NAME"], environ["SERVER_PORT"])
-                path = (
-                    (environ.get("SCRIPT_NAME", "") + environ.get("PATH_INFO", ""))
-                    .encode("latin1")
-                    .decode("utf8")
-                )
-                query_string = environ.get("QUERY_STRING", "").encode("latin-1")
-                host_header = environ.get("HTTP_HOST", None)
-
-            if host_header is not None:
-                url = f"{scheme}://{host_header}{path}"
-            elif server is None:
-                url = path
-            else:
-                host, port = server
-                default_port = {"http": 80, "https": 443, "ws": 80, "wss": 443}[scheme]
-                if port == default_port:
-                    url = f"{scheme}://{host}{path}"
-                else:
-                    url = f"{scheme}://{host}:{port}{path}"
-
-            if query_string:
-                url += "?" + query_string.decode()
-        elif components:
+        if components:
             assert not url, 'Cannot set both "url" and "**components".'
             url = URL("").replace(**components).components.geturl()
+        elif scope is not None:
+            scheme = scope.get("scheme", "http")
+            server = scope.get("server", None)
+            path = scope.get("root_path", "") + scope["path"]
+            query_string = scope.get("query_string", b"")
+
+            host_header = None
+            for key, value in scope["headers"]:
+                if key == b"host":
+                    host_header = value.decode("latin-1")
+                    break
+            url = self._build_url(scheme, path, query_string, server, host_header)
+        elif environ is not None:
+            scheme = environ["wsgi.url_scheme"]
+            server = (environ["SERVER_NAME"], int(environ["SERVER_PORT"]))
+            path = (
+                (environ.get("SCRIPT_NAME", "") + environ.get("PATH_INFO", ""))
+                .encode("latin1")
+                .decode("utf8")
+            )
+            query_string = environ.get("QUERY_STRING", "").encode("latin-1")
+            host_header = environ.get("HTTP_HOST", None)
+            url = self._build_url(scheme, path, query_string, server, host_header)
 
         self._url = url
         self._components = urlsplit(url)
 
+    def _build_url(
+        self,
+        scheme: str,
+        path: str,
+        query_string: bytes = b"",
+        server: typing.Optional[typing.Tuple[str, typing.Optional[int]]] = None,
+        host_header: typing.Optional[str] = None,
+    ) -> str:
+        if host_header is not None:
+            url = f"{scheme}://{host_header}{path}"
+        elif server is None:
+            url = path
+        else:
+            host, port = server
+            default_port = {"http": 80, "https": 443, "ws": 80, "wss": 443}[scheme]
+            if port == default_port or port is None:
+                url = f"{scheme}://{host}{path}"
+            else:
+                url = f"{scheme}://{host}:{port}{path}"
+
+        if query_string:
+            url = f"{url}?{query_string.decode()}"
+
+        return url
+
     @property
     def components(self) -> SplitResult:
         return self._components
@@ -248,17 +357,22 @@ class MultiMapping(typing.Generic[KT, VT], typing.Mapping[KT, VT]):
 
     def __init__(
         self,
-        raw: typing.Union[
-            typing.Mapping[KT, VT],
-            typing.Iterable[typing.Tuple[KT, VT]],
+        raw: typing.Optional[
+            typing.Union[
+                typing.Mapping[KT, VT],
+                typing.Iterable[typing.Tuple[KT, VT]],
+            ]
         ] = None,
     ) -> None:
+        _items: typing.List[typing.Tuple[KT, VT]]
         if raw is None:
             _items = []
         elif isinstance(raw, MultiMapping):
-            _items = list(raw.multi_items())
+            _items = typing.cast(
+                typing.List[typing.Tuple[KT, VT]], list(raw.multi_items())
+            )
         elif isinstance(raw, typing.Mapping):
-            _items = list(raw.items())
+            _items = typing.cast(typing.List[typing.Tuple[KT, VT]], list(raw.items()))
         else:
             _items = list(raw)
 
@@ -274,10 +388,10 @@ class MultiMapping(typing.Generic[KT, VT], typing.Mapping[KT, VT]):
     def __len__(self) -> int:
         return len(self._dict)
 
-    def getlist(self, key: KT) -> typing.Sequence[VT]:
+    def getlist(self, key: KT) -> typing.List[VT]:
         return [item_value for item_key, item_value in self._list if item_key == key]
 
-    def multi_items(self) -> typing.Sequence[typing.Tuple[KT, VT]]:
+    def multi_items(self) -> typing.List[typing.Tuple[KT, VT]]:
         return list(self._list)
 
     def __eq__(self, other: typing.Any) -> bool:
@@ -325,7 +439,7 @@ class MutableMultiMapping(
         elif key in self:
             del self[key]
 
-    def poplist(self, key: KT) -> typing.Sequence[VT]:
+    def poplist(self, key: KT) -> typing.List[VT]:
         values = [v for k, v in self._list if k == key]
         try:
             del self[key]
@@ -347,12 +461,14 @@ class QueryParams(MultiMapping[str, str]):
 
     def __init__(
         self,
-        raw: typing.Union[
-            "MultiMapping[str, str]",
-            typing.Mapping[str, str],
-            typing.Iterable[typing.Tuple[str, str]],
-            str,
-            bytes,
+        raw: typing.Optional[
+            typing.Union[
+                "MultiMapping[str, str]",
+                typing.Mapping[str, str],
+                typing.Iterable[typing.Tuple[str, str]],
+                str,
+                bytes,
+            ]
         ] = None,
     ) -> None:
         if isinstance(raw, str):
@@ -371,119 +487,24 @@ class QueryParams(MultiMapping[str, str]):
         return f"{class_name}({query_string!r})"
 
 
-class UploadFile:
-    """
-    An uploaded file included as part of the request data.
-    """
-
-    __slots__ = ("filename", "content_type", "file")
-
-    spool_max_size = 1024 * 1024
-
-    def __init__(self, filename: str, content_type: str = "") -> None:
-        self.filename = filename
-        self.content_type = content_type
-        self.file = SpooledTemporaryFile(max_size=self.spool_max_size, mode="w+b")
-
-    @property
-    def in_memory(self) -> bool:
-        rolled_to_disk = getattr(self.file, "_rolled", True)
-        return not rolled_to_disk
-
-    def write(self, data: bytes) -> None:
-        self.file.write(data)
-
-    async def awrite(self, data: bytes) -> None:
-        if self.in_memory:
-            self.write(data)
-        else:
-            await asyncio.get_event_loop().run_in_executor(None, self.write, data)
-
-    def read(self, size: int = -1) -> bytes:
-        return self.file.read(size)
-
-    async def aread(self, size: int = -1) -> bytes:
-        if self.in_memory:
-            return self.read(size)
-        return await asyncio.get_event_loop().run_in_executor(None, self.read, size)
-
-    def seek(self, offset: int) -> None:
-        self.file.seek(offset)
-
-    async def aseek(self, offset: int) -> None:
-        if self.in_memory:
-            self.seek(offset)
-        else:
-            await asyncio.get_event_loop().run_in_executor(None, self.seek, offset)
-
-    def close(self) -> None:
-        self.file.close()
-
-    async def aclose(self) -> None:
-        if self.in_memory:
-            self.close()
-        else:
-            await asyncio.get_event_loop().run_in_executor(None, self.close)
-
-    def save(self, filepath: str) -> None:
-        """
-        Save file to disk.
-        """
-        # from shutil.COPY_BUFSIZE
-        copy_bufsize = 1024 * 1024 if os.name == "nt" else 64 * 1024
-        file_position = self.file.tell()
-        self.file.seek(0, 0)
-        try:
-            with open(filepath, "wb+") as target_file:
-                source_read = self.file.read
-                target_write = target_file.write
-                while True:
-                    buf = source_read(copy_bufsize)
-                    if not buf:
-                        break
-                    target_write(buf)
-        finally:
-            self.file.seek(file_position)
-
-    async def asave(self, filepath: str) -> None:
-        """
-        Save file to disk, work in threading pool.
-        """
-        await asyncio.get_event_loop().run_in_executor(None, self.save, filepath)
-
-
-class FormData(MultiMapping[str, typing.Union[str, UploadFile]]):
-    """
-    An immutable MultiMapping, containing both file uploads and text input.
-    """
-
-    __slots__ = MultiMapping.__slots__
-
-    def close(self) -> None:
-        for key, value in self.multi_items():
-            if isinstance(value, UploadFile):
-                value.close()
-
-    async def aclose(self) -> None:
-        for key, value in self.multi_items():
-            if isinstance(value, UploadFile):
-                await value.aclose()
-
-
 class Headers(typing.Mapping[str, str]):
     __slots__ = ("_dict",)
 
     def __init__(
         self,
-        headers: typing.Union[
-            typing.Mapping[str, str],
-            typing.Iterable[typing.Tuple[str, str]],
+        headers: typing.Optional[
+            typing.Union[
+                typing.Mapping[str, str],
+                typing.Iterable[typing.Tuple[str, str]],
+            ]
         ] = None,
     ) -> None:
         store: typing.Dict[str, str] = {}
         items: typing.Iterable[typing.Tuple[str, str]]
         if isinstance(headers, typing.Mapping):
-            items = headers.items()
+            items = typing.cast(
+                typing.Iterable[typing.Tuple[str, str]], headers.items()
+            )
         elif headers is None:
             items = ()
         else:
@@ -522,3 +543,103 @@ class MutableHeaders(Headers, typing.MutableMapping[str, str]):
             self._dict[key] = f"{self._dict[key]}, {value}"
         else:
             self._dict[key] = value
+
+
+class UploadFile:
+    """
+    An uploaded file included as part of the request data.
+    """
+
+    __slots__ = ("filename", "headers", "content_type", "file")
+
+    spool_max_size = 1024 * 1024
+
+    def __init__(self, filename: str, headers: Headers) -> None:
+        self.filename = filename
+        self.headers = headers
+        self.content_type = headers.get("content-type", "")
+        self.file = SpooledTemporaryFile(max_size=self.spool_max_size, mode="w+b")
+
+    @property
+    def in_memory(self) -> bool:
+        rolled_to_disk = getattr(self.file, "_rolled", True)
+        return not rolled_to_disk
+
+    def write(self, data: bytes) -> None:
+        self.file.write(data)
+
+    async def awrite(self, data: bytes) -> None:
+        if self.in_memory:
+            self.write(data)
+        else:
+            await run_in_threadpool(self.write, data)
+
+    def read(self, size: int = -1) -> bytes:
+        return self.file.read(size)
+
+    async def aread(self, size: int = -1) -> bytes:
+        if self.in_memory:
+            return self.read(size)
+        return await run_in_threadpool(self.read, size)
+
+    def seek(self, offset: int) -> None:
+        self.file.seek(offset)
+
+    async def aseek(self, offset: int) -> None:
+        if self.in_memory:
+            self.seek(offset)
+        else:
+            await run_in_threadpool(self.seek, offset)
+
+    def close(self) -> None:
+        self.file.close()
+
+    async def aclose(self) -> None:
+        if self.in_memory:
+            self.close()
+        else:
+            await run_in_threadpool(self.close)
+
+    def save(self, filepath: str) -> None:
+        """
+        Save file to disk.
+        """
+        # from shutil.COPY_BUFSIZE
+        copy_bufsize = 1024 * 1024 if os.name == "nt" else 64 * 1024
+        file_position = self.file.tell()
+        self.file.seek(0, 0)
+        try:
+            with open(filepath, "wb+") as target_file:
+                source_read = self.file.read
+                target_write = target_file.write
+                while True:
+                    buf = source_read(copy_bufsize)
+                    if not buf:
+                        break
+                    target_write(buf)
+        finally:
+            self.file.seek(file_position)
+
+    async def asave(self, filepath: str) -> None:
+        """
+        Save file to disk, work in threading pool.
+        """
+        await run_in_threadpool(self.save, filepath)
+
+
+class FormData(MultiMapping[str, typing.Union[str, UploadFile]]):
+    """
+    An immutable MultiMapping, containing both file uploads and text input.
+    """
+
+    __slots__ = MultiMapping.__slots__
+
+    def close(self) -> None:
+        for key, value in self.multi_items():
+            if isinstance(value, UploadFile):
+                value.close()
+
+    async def aclose(self) -> None:
+        for key, value in self.multi_items():
+            if isinstance(value, UploadFile):
+                await value.aclose()
